@@ -1,4 +1,4 @@
-#     Copyright 2015, Kay Hayen, mailto:kay.hayen@gmail.com
+#     Copyright 2016, Kay Hayen, mailto:kay.hayen@gmail.com
 #
 #     Part of "Nuitka", an optimizing Python compiler that is compatible and
 #     integrates with CPython, but also works on its own.
@@ -23,159 +23,78 @@ source code comments with developer manual sections.
 """
 
 from nuitka.nodes.AssignNodes import (
+    ExpressionTargetTempVariableRef,
+    ExpressionTempVariableRef,
     StatementAssignmentVariable,
     StatementReleaseVariable
 )
-from nuitka.nodes.BuiltinIteratorNodes import (
-    ExpressionBuiltinIter1,
-    ExpressionBuiltinNext1
-)
+from nuitka.nodes.BuiltinIteratorNodes import ExpressionBuiltinIter1
+from nuitka.nodes.BuiltinNextNodes import ExpressionBuiltinNext1
+from nuitka.nodes.CodeObjectSpecs import CodeObjectSpec
 from nuitka.nodes.ConditionalNodes import StatementConditional
-from nuitka.nodes.ConstantRefNodes import ExpressionConstantRef
+from nuitka.nodes.ConstantRefNodes import makeConstantRefNode
 from nuitka.nodes.ContainerOperationNodes import (
-    ExpressionDictOperationSet,
-    ExpressionListOperationAppend,
-    ExpressionSetOperationAdd
+    StatementListOperationAppend,
+    StatementSetOperationAdd
 )
+from nuitka.nodes.DictionaryNodes import StatementDictOperationSet
+from nuitka.nodes.FrameNodes import StatementsFrame
 from nuitka.nodes.FunctionNodes import (
     ExpressionFunctionBody,
     ExpressionFunctionCall,
     ExpressionFunctionCreation,
     ExpressionFunctionRef
 )
-from nuitka.nodes.LoopNodes import StatementBreakLoop, StatementLoop
+from nuitka.nodes.GeneratorNodes import (
+    ExpressionGeneratorObjectBody,
+    ExpressionMakeGeneratorObject
+)
+from nuitka.nodes.LoopNodes import StatementLoop, StatementLoopBreak
+from nuitka.nodes.NodeMakingHelpers import (
+    makeVariableRefNode,
+    makeVariableTargetRefNode
+)
+from nuitka.nodes.OutlineNodes import ExpressionOutlineBody
 from nuitka.nodes.ParameterSpecs import ParameterSpec
 from nuitka.nodes.ReturnNodes import StatementReturn
 from nuitka.nodes.StatementNodes import (
     StatementExpressionOnly,
     StatementGeneratorEntry,
-    StatementsFrame,
     StatementsSequence
 )
-from nuitka.nodes.VariableRefNodes import (
-    ExpressionTargetTempVariableRef,
-    ExpressionTempVariableRef,
-    ExpressionVariableRef
-)
 from nuitka.nodes.YieldNodes import ExpressionYield
-from nuitka.utils import Utils
+from nuitka.PythonVersions import python_version
 
 from .Helpers import (
     buildNode,
     buildNodeList,
     getKind,
     makeStatementsSequenceFromStatement,
-    makeTryFinallyExpression,
-    makeTryFinallyStatement,
-    mergeStatements,
-    wrapTryFinallyLater
+    mergeStatements
 )
 from .ReformulationAssignmentStatements import buildAssignmentStatements
 from .ReformulationBooleanExpressions import buildAndNode
 from .ReformulationTryExceptStatements import makeTryExceptSingleHandlerNode
+from .ReformulationTryFinallyStatements import makeTryFinallyStatement
 
 
 def buildListContractionNode(provider, node, source_ref):
     # List contractions are dealt with by general code.
 
-    if Utils.python_version < 300:
-        temp_scope = provider.allocateTempScope("listcontr")
-
-        outer_iter_var = provider.allocateTempVariable(
-            temp_scope = temp_scope,
-            name       = "listcontr_iter"
-        )
-
-        outer_iter_ref = ExpressionTempVariableRef(
-            variable   = outer_iter_var,
-            source_ref = source_ref
-        )
-
-        container_tmp   = provider.allocateTempVariable(
-            temp_scope = temp_scope,
-            name       = "listcontr_result"
-        )
-
-        statements, release_statements = _buildContractionBodyNode(
-            provider        = provider,
-            node            = node,
-            emit_class      = ExpressionListOperationAppend,
-            start_value     = ExpressionConstantRef(
-                constant   = [],
-                source_ref = source_ref
-            ),
-            outer_iter_ref  = outer_iter_ref,
-            container_tmp   = container_tmp,
-            temp_scope      = temp_scope,
-            assign_provider = True,
-            source_ref      = source_ref,
-            function_body   = provider
-        )
-
-        statements.insert(
-            0,
-            StatementAssignmentVariable(
-                variable_ref = ExpressionTargetTempVariableRef(
-                    variable   = outer_iter_var,
-                    source_ref = source_ref
-                ),
-                source       = ExpressionBuiltinIter1(
-                    value      = buildNode(
-                        provider   = provider,
-                        node       = node.generators[0].iter,
-                        source_ref = source_ref
-                    ),
-                    source_ref = source_ref
-                ),
-                source_ref   = source_ref
-            )
-        )
-
-        result = makeTryFinallyExpression(
-            expression = ExpressionTempVariableRef(
-                variable   = container_tmp,
-                source_ref = source_ref
-            ),
-            tried      = statements,
-            final      = release_statements,
-            source_ref = source_ref
-        )
-
-        final = StatementsSequence(
-            statements = (
-                StatementReleaseVariable(
-                    variable   = container_tmp,
-                    tolerant   = True,
-                    source_ref = source_ref
-                ),
-                StatementReleaseVariable(
-                    variable   = outer_iter_var,
-                    tolerant   = True,
-                    source_ref = source_ref
-                ),
-            ),
-            source_ref = source_ref
-        )
-
-        wrapTryFinallyLater(
-            node  = result,
-            final = final
-        )
-
-        return result
-
     return _buildContractionNode(
         provider        = provider,
         node            = node,
-        name            = "<listcontraction>",
-        emit_class      = ExpressionListOperationAppend,
-        start_value     = ExpressionConstantRef(
+        name            = "list_contraction"
+                            if python_version < 300 else
+                          "<listcontraction>",
+        emit_class      = StatementListOperationAppend,
+        start_value     = makeConstantRefNode(
             constant   = [],
             source_ref = source_ref
         ),
         # Note: For Python3, the list contractions no longer assign to the outer
         # scope.
-        assign_provider = Utils.python_version < 300,
+        assign_provider = python_version < 300,
         source_ref      = source_ref
     )
 
@@ -187,8 +106,8 @@ def buildSetContractionNode(provider, node, source_ref):
         provider        = provider,
         node            = node,
         name            = "<setcontraction>",
-        emit_class      = ExpressionSetOperationAdd,
-        start_value     = ExpressionConstantRef(
+        emit_class      = StatementSetOperationAdd,
+        start_value     = makeConstantRefNode(
             constant   = set(),
             source_ref = source_ref
         ),
@@ -204,8 +123,8 @@ def buildDictContractionNode(provider, node, source_ref):
         provider        = provider,
         node            = node,
         name            = "<dictcontraction>",
-        emit_class      = ExpressionDictOperationSet,
-        start_value     = ExpressionConstantRef(
+        emit_class      = StatementDictOperationSet,
+        start_value     = makeConstantRefNode(
             constant   = {},
             source_ref = source_ref
         ),
@@ -231,13 +150,44 @@ def buildGeneratorExpressionNode(provider, node, source_ref):
 
 
 def _buildContractionBodyNode(provider, node, emit_class, start_value,
-                              container_tmp, outer_iter_ref, temp_scope,
-                              assign_provider, source_ref, function_body):
+                              container_tmp, iter_tmp, temp_scope,
+                              assign_provider, function_body, source_ref):
+
     # This uses lots of variables and branches. There is no good way
     # around that, and we deal with many cases, due to having generator
     # expressions sharing this code, pylint: disable=R0912,R0914
-    if start_value is not None:
+    tmp_variables = []
+
+    if assign_provider:
+        tmp_variables.append(iter_tmp)
+
+    if container_tmp is not None:
+        tmp_variables.append(container_tmp)
+
+    # First assign the iterator if we are an outline.
+    if assign_provider:
         statements = [
+            StatementAssignmentVariable(
+                variable_ref = makeVariableTargetRefNode(
+                    variable   = iter_tmp,
+                    source_ref = source_ref
+                ),
+                source       = ExpressionBuiltinIter1(
+                    value      = buildNode(
+                        provider   = provider,
+                        node       = node.generators[0].iter,
+                        source_ref = source_ref
+                    ),
+                    source_ref = source_ref
+                ),
+                source_ref   = source_ref.atInternal()
+            )
+        ]
+    else:
+        statements = []
+
+    if start_value is not None:
+        statements.append(
             StatementAssignmentVariable(
                 variable_ref = ExpressionTargetTempVariableRef(
                     variable   = container_tmp,
@@ -246,15 +196,8 @@ def _buildContractionBodyNode(provider, node, emit_class, start_value,
                 source       = start_value,
                 source_ref   = source_ref.atInternal()
             )
-        ]
+        )
 
-        if assign_provider:
-            tmp_variables = []
-        else:
-            tmp_variables = [container_tmp]
-    else:
-        statements = []
-        tmp_variables = []
 
     if hasattr(node, "elt"):
         if start_value is not None:
@@ -273,8 +216,6 @@ def _buildContractionBodyNode(provider, node, emit_class, start_value,
         else:
             assert emit_class is ExpressionYield
 
-            function_body.markAsGenerator()
-
             current_body = emit_class(
                 buildNode(
                     provider   = function_body,
@@ -284,10 +225,10 @@ def _buildContractionBodyNode(provider, node, emit_class, start_value,
                 source_ref = source_ref
             )
     else:
-        assert emit_class is ExpressionDictOperationSet
+        assert emit_class is StatementDictOperationSet
 
-        current_body = emit_class(
-            ExpressionTempVariableRef(
+        current_body = StatementDictOperationSet(
+            dict_arg   = ExpressionTempVariableRef(
                 variable   = container_tmp,
                 source_ref = source_ref
             ),
@@ -304,10 +245,11 @@ def _buildContractionBodyNode(provider, node, emit_class, start_value,
             source_ref = source_ref
         )
 
-    current_body = StatementExpressionOnly(
-        expression = current_body,
-        source_ref = source_ref
-    )
+    if current_body.isExpression():
+        current_body = StatementExpressionOnly(
+            expression = current_body,
+            source_ref = source_ref
+        )
 
     for count, qual in enumerate(reversed(node.generators)):
         tmp_value_variable = function_body.allocateTempVariable(
@@ -322,7 +264,10 @@ def _buildContractionBodyNode(provider, node, emit_class, start_value,
         # inside the function.
 
         if qual is node.generators[0]:
-            iterator_ref = outer_iter_ref.makeCloneAt(source_ref)
+            iterator_ref = makeVariableRefNode(
+                variable   = iter_tmp,
+                source_ref = source_ref
+            )
 
             tmp_iter_variable = None
 
@@ -363,26 +308,21 @@ def _buildContractionBodyNode(provider, node, emit_class, start_value,
 
         loop_statements = [
             makeTryExceptSingleHandlerNode(
-                tried          = makeStatementsSequenceFromStatement(
-                    statement = StatementAssignmentVariable(
-                        variable_ref = ExpressionTargetTempVariableRef(
-                            variable   = tmp_value_variable,
-                            source_ref = source_ref
-                        ),
-                        source       = ExpressionBuiltinNext1(
-                            value      = iterator_ref,
-                            source_ref = source_ref
-                        ),
-                        source_ref   = source_ref
-                    )
+                tried          = StatementAssignmentVariable(
+                    variable_ref = ExpressionTargetTempVariableRef(
+                        variable   = tmp_value_variable,
+                        source_ref = source_ref
+                    ),
+                    source       = ExpressionBuiltinNext1(
+                        value      = iterator_ref,
+                        source_ref = source_ref
+                    ),
+                    source_ref   = source_ref
                 ),
                 exception_name = "StopIteration",
-                handler_body   = makeStatementsSequenceFromStatement(
-                    statement = StatementBreakLoop(
-                        source_ref = source_ref.atInternal()
-                    )
+                handler_body   = StatementLoopBreak(
+                    source_ref = source_ref.atInternal()
                 ),
-                public_exc     = False,
                 source_ref     = source_ref
             ),
             buildAssignmentStatements(
@@ -403,22 +343,10 @@ def _buildContractionBodyNode(provider, node, emit_class, start_value,
             source_ref = source_ref
         )
 
-        if len(conditions) == 1:
-            loop_statements.append(
-                StatementConditional(
-                    condition  = conditions[0],
-                    yes_branch = makeStatementsSequenceFromStatement(
-                        statement = current_body
-                    ),
-                    no_branch  = None,
-                    source_ref = source_ref
-                )
-            )
-        elif len(conditions) > 1:
+        if len(conditions) >= 1:
             loop_statements.append(
                 StatementConditional(
                     condition  = buildAndNode(
-                        provider   = function_body,
                         values     = conditions,
                         source_ref = source_ref
                     ),
@@ -446,13 +374,12 @@ def _buildContractionBodyNode(provider, node, emit_class, start_value,
             nested_statements.append(
                 StatementReleaseVariable(
                     variable   = tmp_iter_variable,
-                    tolerant   = False,
                     source_ref = source_ref
                 )
             )
 
         current_body = StatementsSequence(
-            statements = nested_statements,
+            statements = mergeStatements(nested_statements, False),
             source_ref = source_ref
         )
 
@@ -470,7 +397,6 @@ def _buildContractionBodyNode(provider, node, emit_class, start_value,
     release_statements = [
         StatementReleaseVariable(
             variable   = tmp_variable,
-            tolerant   = True,
             source_ref = source_ref
         )
         for tmp_variable in
@@ -488,46 +414,97 @@ def _buildContractionNode(provider, node, name, emit_class, start_value,
 
     # Note: The assign_provider is only to cover Python2 list contractions,
     # assigning one of the loop variables to the outside scope.
+    if assign_provider:
+        function_body = ExpressionOutlineBody(
+            provider   = provider,
+            name       = name,
+            source_ref = source_ref
+        )
 
-    assert provider.isParentVariableProvider(), provider
+        iter_tmp = function_body.allocateTempVariable(
+            temp_scope = None,
+            name       = ".0"
+        )
+    else:
+        # TODO: No function ought to be necessary.
 
-    iterator_var_name = ".0"
+        function_body = ExpressionFunctionBody(
+            provider   = provider,
+            name       = name,
+            doc        = None,
+            parameters = ParameterSpec(
+                ps_name          = name,
+                ps_normal_args   = (".0",),
+                ps_list_star_arg = None,
+                ps_dict_star_arg = None,
+                ps_default_count = 0,
+                ps_kw_only_args  = ()
+            ),
+            flags      = set(),
+            source_ref = source_ref
+        )
 
-    function_body = ExpressionFunctionBody(
-        provider   = provider,
-        name       = name,
-        doc        = None,
-        parameters = ParameterSpec(
-            name          = "contraction",
-            normal_args   = (iterator_var_name,),
-            list_star_arg = None,
-            dict_star_arg = None,
-            default_count = 0,
-            kw_only_args  = ()
-        ),
-        source_ref = source_ref
+        iter_tmp = function_body.getVariableForAssignment(
+            variable_name = ".0"
+        )
+        assert iter_tmp.isParameterVariable()
+
+    code_object = CodeObjectSpec(
+        co_name           = name,
+        co_kind           = "Generator" if emit_class is ExpressionYield else "Function",
+        co_varnames       = () if assign_provider else function_body.getParameters().getParameterNames(),
+        co_argcount       = 0 if assign_provider else len(function_body.getParameters().getParameterNames()),
+        co_kwonlyargcount = 0,
+        co_has_starlist   = False,
+        co_has_stardict   = False,
     )
 
+    if emit_class is ExpressionYield:
+        code_body = ExpressionGeneratorObjectBody(
+            provider   = function_body,
+            name       = "<genexpr>",
+            flags      = set(),
+            source_ref = source_ref
+        )
+
+        iter_tmp = code_body.getVariableForReference(
+            variable_name = ".0"
+        )
+        assert iter_tmp.isLocalVariable()
+
+        function_body.setBody(
+            makeStatementsSequenceFromStatement(
+                statement = StatementReturn(
+                    expression = ExpressionMakeGeneratorObject(
+                        generator_ref = ExpressionFunctionRef(
+                            function_body = code_body,
+                            source_ref    = source_ref
+                        ),
+                        code_object   = code_object,
+                        source_ref    = source_ref
+                    ),
+                    source_ref = source_ref
+                )
+            )
+        )
+    else:
+        code_body = function_body
+
     if start_value is not None:
-        container_tmp = function_body.allocateTempVariable(
+        container_tmp = code_body.allocateTempVariable(
             temp_scope = None,
             name       = "contraction_result"
         )
     else:
         container_tmp = None
 
-    outer_iter_ref = ExpressionVariableRef(
-        variable_name = iterator_var_name,
-        source_ref    = source_ref
-    )
-
-    statements, del_statements = _buildContractionBodyNode(
-        function_body   = function_body,
+    statements, release_statements = _buildContractionBodyNode(
+        function_body   = code_body,
         assign_provider = assign_provider,
         provider        = provider,
         node            = node,
         emit_class      = emit_class,
-        outer_iter_ref  = outer_iter_ref,
+        iter_tmp        = iter_tmp,
         temp_scope      = None,
         start_value     = start_value,
         container_tmp   = container_tmp,
@@ -547,50 +524,50 @@ def _buildContractionNode(provider, node, name, emit_class, start_value,
 
     statements = (
         makeTryFinallyStatement(
+            provider   = function_body,
             tried      = statements,
-            final      = del_statements,
+            final      = release_statements,
             source_ref = source_ref.atInternal()
         ),
     )
 
-    function_body.setBody(
+    code_body.setBody(
         makeStatementsSequenceFromStatement(
             statement = StatementsFrame(
-                statements    = statements,
-                guard_mode    = "pass_through"
+                statements  = mergeStatements(statements, False),
+                guard_mode  = "pass_through"
                                   if emit_class is not ExpressionYield else
                                 "generator",
-                var_names     = (),
-                arg_count     = 0,
-                kw_only_count = 0,
-                has_starlist  = False,
-                has_stardict  = False,
-                code_name     = "contraction",
-                source_ref    = source_ref
+                code_object = code_object,
+                source_ref  = source_ref
             )
         )
     )
 
-    return ExpressionFunctionCall(
-        function   = ExpressionFunctionCreation(
-            function_ref = ExpressionFunctionRef(
-                function_body = function_body,
-                source_ref    = source_ref
+    if not assign_provider:
+        return ExpressionFunctionCall(
+            function   = ExpressionFunctionCreation(
+                function_ref = ExpressionFunctionRef(
+                    function_body = function_body,
+                    source_ref    = source_ref
+                ),
+                code_object  = code_object,
+                defaults     = (),
+                kw_defaults  = None,
+                annotations  = None,
+                source_ref   = source_ref
             ),
-            defaults     = (),
-            kw_defaults  = None,
-            annotations  = None,
-            source_ref   = source_ref
-        ),
-        values     = (
-            ExpressionBuiltinIter1(
-                value      = buildNode(
-                    provider   = provider,
-                    node       = node.generators[0].iter,
+            values     = (
+                ExpressionBuiltinIter1(
+                    value      = buildNode(
+                        provider   = provider,
+                        node       = node.generators[0].iter,
+                        source_ref = source_ref
+                    ),
                     source_ref = source_ref
                 ),
-                source_ref = source_ref
             ),
-        ),
-        source_ref = source_ref
-    )
+            source_ref = source_ref
+        )
+    else:
+        return function_body
